/*
 * Open Surge Engine
 * confirmbox.h - confirm box
 * Copyright 2008-2026 Alexandre Martins <alemartf(at)gmail.com>
 * http://opensurge2d.org
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _CONFIRMBOX_H
#define _CONFIRMBOX_H

/* confirm box data structure */
typedef struct confirmboxdata_t {
    const char* message;
    const char* option1;
    const char* option2; /* set it to NULL if you want only one option */
    int default_option;  /* default option: 1 or 2 */
} confirmboxdata_t;

/* public functions */
void confirmbox_init(void* confirmbox);
void confirmbox_release();
void confirmbox_update();
void confirmbox_render();

/* interface */
int confirmbox_selected_option(); /* returns 1 or 2 (or 0 if nothing has been selected) */

#endif
