// -----------------------------------------------------------------------------
// File: ui_settings.ss
// Description: UI Settings plugin (Debug Mode)
// Author: Alexandre Martins <http://opensurge2d.org>
// License: MIT
// -----------------------------------------------------------------------------

/*

This plugin has UI settings that are useful to other plugins.

*/

object "Debug Mode - UI Settings" is "debug-mode-plugin"
{
    public readonly zindex = 10000;
}